@@ -33,6 +33,7 @@ import Data.Char
 import Data.List
 import Data.Maybe
 import qualified Data.Map as Map
+import qualified Data.Set as Set
 import Test.QuickCheck.All (forAllProperties)
 import Test.QuickCheck.Test (quickCheckWithResult, stdArgs, maxSuccess)
 
@@ -136,6 +137,46 @@ prop_checkBashisms53= verifyNot checkBashisms "#!/bin/sh\nprintf -- -f\n"
 prop_checkBashisms54= verify checkBashisms "#!/bin/sh\nfoo+=bar"
 prop_checkBashisms55= verify checkBashisms "#!/bin/sh\necho ${@%foo}"
 prop_checkBashisms56= verifyNot checkBashisms "#!/bin/sh\necho ${##}"
+prop_checkBashisms57= verifyNot checkBashisms "#!/bin/dash\nulimit -c 0"
+prop_checkBashisms58= verify checkBashisms "#!/bin/sh\nulimit -c 0"
+prop_checkBashisms59 = verify checkBashisms "#!/bin/sh\njobs -s"
+prop_checkBashisms60 = verifyNot checkBashisms "#!/bin/sh\njobs -p"
+prop_checkBashisms61 = verifyNot checkBashisms "#!/bin/sh\njobs -lp"
+prop_checkBashisms62 = verify checkBashisms "#!/bin/sh\nexport -f foo"
+prop_checkBashisms63 = verifyNot checkBashisms "#!/bin/sh\nexport -p"
+prop_checkBashisms64 = verify checkBashisms "#!/bin/sh\nreadonly -a"
+prop_checkBashisms65 = verifyNot checkBashisms "#!/bin/sh\nreadonly -p"
+prop_checkBashisms66 = verifyNot checkBashisms "#!/bin/sh\ncd -P ."
+prop_checkBashisms67 = verify checkBashisms "#!/bin/sh\ncd -P -e ."
+prop_checkBashisms68 = verify checkBashisms "#!/bin/sh\numask -p"
+prop_checkBashisms69 = verifyNot checkBashisms "#!/bin/sh\numask -S"
+prop_checkBashisms70 = verify checkBashisms "#!/bin/sh\ntrap -l"
+prop_checkBashisms71 = verify checkBashisms "#!/bin/sh\ntype -a ls"
+prop_checkBashisms72 = verifyNot checkBashisms "#!/bin/sh\ntype ls"
+prop_checkBashisms73 = verify checkBashisms "#!/bin/sh\nunset -n namevar"
+prop_checkBashisms74 = verifyNot checkBashisms "#!/bin/sh\nunset -f namevar"
+prop_checkBashisms75 = verifyNot checkBashisms "#!/bin/sh\necho \"-n foo\""
+prop_checkBashisms76 = verifyNot checkBashisms "#!/bin/sh\necho \"-ne foo\""
+prop_checkBashisms77 = verifyNot checkBashisms "#!/bin/sh\necho -Q foo"
+prop_checkBashisms78 = verify checkBashisms "#!/bin/sh\necho -ne foo"
+prop_checkBashisms79 = verify checkBashisms "#!/bin/sh\nhash -l"
+prop_checkBashisms80 = verifyNot checkBashisms "#!/bin/sh\nhash -r"
+prop_checkBashisms81 = verifyNot checkBashisms "#!/bin/dash\nhash -v"
+prop_checkBashisms82 = verifyNot checkBashisms "#!/bin/sh\nset -v +o allexport -o errexit -C"
+prop_checkBashisms83 = verifyNot checkBashisms "#!/bin/sh\nset --"
+prop_checkBashisms84 = verify checkBashisms "#!/bin/sh\nset -o pipefail"
+prop_checkBashisms85 = verify checkBashisms "#!/bin/sh\nset -B"
+prop_checkBashisms86 = verifyNot checkBashisms "#!/bin/dash\nset -o emacs"
+prop_checkBashisms87 = verify checkBashisms "#!/bin/sh\nset -o emacs"
+prop_checkBashisms88 = verifyNot checkBashisms "#!/bin/sh\nset -- wget -o foo 'https://some.url'"
+prop_checkBashisms89 = verifyNot checkBashisms "#!/bin/sh\nopts=$-\nset -\"$opts\""
+prop_checkBashisms90 = verifyNot checkBashisms "#!/bin/sh\nset -o \"$opt\""
+prop_checkBashisms91 = verify checkBashisms "#!/bin/sh\nwait -n"
+prop_checkBashisms92 = verify checkBashisms "#!/bin/sh\necho $((16#FF))"
+prop_checkBashisms93 = verify checkBashisms "#!/bin/sh\necho $(( 10#$(date +%m) ))"
+prop_checkBashisms94 = verify checkBashisms "#!/bin/sh\n[ -v var ]"
+prop_checkBashisms95 = verify checkBashisms "#!/bin/sh\necho $_"
+prop_checkBashisms96 = verifyNot checkBashisms "#!/bin/dash\necho $_"
 checkBashisms = ForShell [Sh, Dash] $ \t -> do
     params <- ask
     kludge params t
@@ -170,6 +211,8 @@ checkBashisms = ForShell [Sh, Dash] $ \t -> do
             warnMsg id "== in place of = is"
     bashism (TC_Binary id SingleBracket "=~" _ _) =
             warnMsg id "=~ regex matching is"
+    bashism (TC_Unary id SingleBracket "-v" _) =
+            warnMsg id "unary -v (in place of [ -n \"${var+x}\" ]) is"
     bashism (TC_Unary id _ "-a" _) =
             warnMsg id "unary -a in place of -e is"
     bashism (TA_Unary id op _)
@@ -194,7 +237,7 @@ checkBashisms = ForShell [Sh, Dash] $ \t -> do
     bashism t@(TA_Variable id str _) | isBashVariable str =
         warnMsg id $ str ++ " is"
 
-    bashism t@(T_DollarBraced id token) = do
+    bashism t@(T_DollarBraced id _ token) = do
         mapM_ check expansion
         when (isBashVariable var) $
                     warnMsg id $ var ++ " is"
@@ -222,20 +265,71 @@ checkBashisms = ForShell [Sh, Dash] $ \t -> do
         warnMsg id "`<file` to read files is"
 
     bashism t@(T_SimpleCommand _ _ (cmd:arg:_))
-        | t `isCommand` "echo" && "-" `isPrefixOf` argString =
-            unless ("--" `isPrefixOf` argString) $ -- echo "-----"
-                if isDash
-                then
-                    when (argString /= "-n") $
-                        warnMsg (getId arg) "echo flags besides -n"
-                else
-                    warnMsg (getId arg) "echo flags are"
-      where argString = concat $ oversimplify arg
+        | t `isCommand` "echo" && argString `matches` flagRegex =
+            if isDash
+            then
+                when (argString /= "-n") $
+                    warnMsg (getId arg) "echo flags besides -n"
+            else
+                warnMsg (getId arg) "echo flags are"
+      where
+          argString = concat $ oversimplify arg
+          flagRegex = mkRegex "^-[eEsn]+$"
+
     bashism t@(T_SimpleCommand _ _ (cmd:arg:_))
         | t `isCommand` "exec" && "-" `isPrefixOf` concat (oversimplify arg) =
             warnMsg (getId arg) "exec flags are"
     bashism t@(T_SimpleCommand id _ _)
         | t `isCommand` "let" = warnMsg id "'let' is"
+    bashism t@(T_SimpleCommand _ _ (cmd:args))
+        | t `isCommand` "set" = unless isDash $
+            checkOptions $ getLiteralArgs args
+      where
+        -- Get the literal options from a list of arguments,
+        -- up until the first non-literal one
+        getLiteralArgs :: [Token] -> [(Id, String)]
+        getLiteralArgs (first:rest) = fromMaybe [] $ do
+            str <- getLiteralString first
+            return $ (getId first, str) : getLiteralArgs rest
+        getLiteralArgs [] = []
+
+        -- Check a flag-option pair (such as -o errexit)
+        checkOptions (flag@(fid,flag') : opt@(oid,opt') : rest)
+            | flag' `matches` oFlagRegex = do
+                when (opt' `notElem` longOptions) $
+                  warnMsg oid $ "set option " <> opt' <> " is"
+                checkFlags (flag:rest)
+            | otherwise = checkFlags (flag:opt:rest)
+        checkOptions (flag:rest) = checkFlags (flag:rest)
+        checkOptions _           = return ()
+
+        -- Check that each option in a sequence of flags
+        -- (such as -aveo) is valid
+        checkFlags (flag@(fid, flag'):rest)
+            | startsOption flag' = do
+                unless (flag' `matches` validFlagsRegex) $
+                  forM_ (tail flag') $ \letter ->
+                    when (letter `notElem` optionsSet) $
+                      warnMsg fid $ "set flag " <> ('-':letter:" is")
+                checkOptions rest
+            | beginsWithDoubleDash flag' = do
+                warnMsg fid $ "set flag " <> flag' <> " is"
+                checkOptions rest
+            -- Either a word that doesn't start with a dash, or simply '--',
+            -- so stop checking.
+            | otherwise = return ()
+        checkFlags [] = return ()
+
+        options              = "abCefhmnuvxo"
+        optionsSet           = Set.fromList options
+        startsOption         = (`matches` mkRegex "^(\\+|-[^-])")
+        oFlagRegex           = mkRegex $ "^[-+][" <> options <> "]*o$"
+        validFlagsRegex      = mkRegex $ "^[-+]([" <> options <> "]+o?|o)$"
+        beginsWithDoubleDash = (`matches` mkRegex "^--.+$")
+        longOptions          = Set.fromList
+            [ "allexport", "errexit", "ignoreeof", "monitor", "noclobber"
+            , "noexec", "noglob", "nolog", "notify" , "nounset", "verbose"
+            , "vi", "xtrace" ]
 
     bashism t@(T_SimpleCommand id _ (cmd:rest)) =
         let name = fromMaybe "" $ getCommandName t
@@ -244,7 +338,8 @@ checkBashisms = ForShell [Sh, Dash] $ \t -> do
             when (name `elem` unsupportedCommands) $
                 warnMsg id $ "'" ++ name ++ "' is"
             potentially $ do
-                allowed <- Map.lookup name allowedFlags
+                allowed' <- Map.lookup name allowedFlags
+                allowed <- allowed'
                 (word, flag) <- listToMaybe $
                     filter (\x -> (not . null . snd $ x) && snd x `notElem` allowed) flags
                 return . warnMsg (getId word) $ name ++ " -" ++ flag ++ " is"
@@ -279,16 +374,28 @@ checkBashisms = ForShell [Sh, Dash] $ \t -> do
             "typeset"
             ] ++ if not isDash then ["local"] else []
         allowedFlags = Map.fromList [
-            ("exec", []),
-            ("export", ["-p"]),
-            ("printf", []),
-            ("read", if isDash then ["r", "p"] else ["r"]),
-            ("ulimit", ["f"])
+            ("cd", Just ["L", "P"]),
+            ("exec", Just []),
+            ("export", Just ["p"]),
+            ("hash", Just $ if isDash then ["r", "v"] else ["r"]),
+            ("jobs", Just ["l", "p"]),
+            ("printf", Just []),
+            ("read", Just $ if isDash then ["r", "p"] else ["r"]),
+            ("readonly", Just ["p"]),
+            ("trap", Just []),
+            ("type", Just []),
+            ("ulimit", if isDash then Nothing else Just ["f"]),
+            ("umask", Just ["S"]),
+            ("unset", Just ["f", "v"]),
+            ("wait", Just [])
             ]
     bashism t@(T_SourceCommand id src _) =
         let name = fromMaybe "" $ getCommandName src
-        in do
-            when (name == "source") $ warnMsg id "'source' in place of '.' is"
+        in when (name == "source") $ warnMsg id "'source' in place of '.' is"
+    bashism (TA_Expansion _ (T_Literal id str : _)) | str `matches` radix =
+        when (str `matches` radix) $ warnMsg id "arithmetic base conversion is"
+      where
+        radix = mkRegex "^[0-9]+#"
     bashism _ = return ()
 
     varChars="_0-9a-zA-Z"
@@ -303,10 +410,11 @@ checkBashisms = ForShell [Sh, Dash] $ \t -> do
         ]
     bashVars = [
         "OSTYPE", "MACHTYPE", "HOSTTYPE", "HOSTNAME",
-        "DIRSTACK", "EUID", "UID", "SHLVL", "PIPESTATUS", "SHELLOPTS"
+        "DIRSTACK", "EUID", "UID", "SHLVL", "PIPESTATUS", "SHELLOPTS",
+        "_"
         ]
     bashDynamicVars = [ "RANDOM", "SECONDS" ]
-    dashVars = [ ]
+    dashVars = [ "_" ]
     isBashVariable var =
         (var `elem` bashDynamicVars
             || var `elem` bashVars && not (isAssigned var))
@@ -318,31 +426,45 @@ checkBashisms = ForShell [Sh, Dash] $ \t -> do
                 _ -> False
 
 prop_checkEchoSed1 = verify checkEchoSed "FOO=$(echo \"$cow\" | sed 's/foo/bar/g')"
+prop_checkEchoSed1b= verify checkEchoSed "FOO=$(sed 's/foo/bar/g' <<< \"$cow\")"
 prop_checkEchoSed2 = verify checkEchoSed "rm $(echo $cow | sed -e 's,foo,bar,')"
+prop_checkEchoSed2b= verify checkEchoSed "rm $(sed -e 's,foo,bar,' <<< $cow)"
 checkEchoSed = ForShell [Bash, Ksh] f
   where
+    f (T_Redirecting id lefts r) =
+        when (any redirectHereString lefts) $
+            checkSed id rcmd
+      where
+        redirectHereString :: Token -> Bool
+        redirectHereString t = case t of
+            (T_FdRedirect _ _ T_HereString{}) -> True
+            _                                 -> False
+        rcmd = oversimplify r
+
     f (T_Pipeline id _ [a, b]) =
         when (acmd == ["echo", "${VAR}"]) $
-            case bcmd of
-                ["sed", v] -> checkIn v
-                ["sed", "-e", v] -> checkIn v
-                _ -> return ()
+            checkSed id bcmd
       where
-        -- This should have used backreferences, but TDFA doesn't support them
-        sedRe = mkRegex "^s(.)([^\n]*)g?$"
-        isSimpleSed s = fromMaybe False $ do
-            [first,rest] <- matchRegex sedRe s
-            let delimiters = filter (== head first) rest
-            guard $ length delimiters == 2
-            return True
-
         acmd = oversimplify a
         bcmd = oversimplify b
-        checkIn s =
-            when (isSimpleSed s) $
-                style id 2001 "See if you can use ${variable//search/replace} instead."
+
     f _ = return ()
 
+    checkSed id ["sed", v]       = checkIn id v
+    checkSed id ["sed", "-e", v] = checkIn id v
+    checkSed _ _                 = return ()
+
+    -- This should have used backreferences, but TDFA doesn't support them
+    sedRe = mkRegex "^s(.)([^\n]*)g?$"
+    isSimpleSed s = fromMaybe False $ do
+        [first,rest] <- matchRegex sedRe s
+        let delimiters = filter (== head first) rest
+        guard $ length delimiters == 2
+        return True
+    checkIn id s =
+        when (isSimpleSed s) $
+            style id 2001 "See if you can use ${variable//search/replace} instead."
+
 
 prop_checkBraceExpansionVars1 = verify checkBraceExpansionVars "echo {1..$n}"
 prop_checkBraceExpansionVars2 = verifyNot checkBraceExpansionVars "echo {1,3,$n}"
